/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl;

import java.util.List;
import java.util.Map;

import org.xmloperator.lambda3.exception.term.BrokenTermException;
import org.xmloperator.lambda3.exception.term.StillReferencedTermException;
import org.xmloperator.lambda3.impl.type.TermTypeFactory;
import org.xmloperator.lambda3.model.Constant;
import org.xmloperator.lambda3.model.Function2;
import org.xmloperator.lambda3.model.Term;
import org.xmloperator.lambda3.model.TermRecycler;
import org.xmloperator.lambda3.model.Variable;
import org.xmloperator.lambda3.model.type.TermType;
import org.xmloperator.lambda3.model.type.TermTypeFormula;
import org.xmloperator.lambda3.util.Stack;

/**
 * Implementation of a Constant relative to a function with two arguments.
 */
final class Constant2Impl extends ConstantImpl<Function2> {
	static int creationCount = 0;
	static int breakCount = 0;
	static final Stack<Constant2Impl> reusableConstantStack = new Stack<Constant2Impl>(
			Constant2Impl.class, REUSE_CAPACITY);

	/**
	 * Private constructor.
	 */
	private static final Constant2Impl newConstant(Function2 function) {
		if (function == null)
			throw new IllegalArgumentException();
		Constant2Impl constant = reusableConstantStack.pop();
		if (constant == null)
			constant = new Constant2Impl();
		constant.function = function;
		creationCount++;
		return constant;
	}

	/**
	 * Constructor for integer operators.
	 */
	static final Constant2Impl newConstant(Function2 function,
			TermType termType0, TermType termType1) {
		if (termType0 == null || termType1 == null)
			throw new IllegalArgumentException();
		Constant2Impl constant = newConstant(function);
		constant.termType0 = termType0;
		constant.termTypeFormula0 = null;
		constant.termType1 = termType1;
		return constant;
	}

	/**
	 * Constructor for Constant T.
	 */
	static final Constant2Impl newConstant(Function2 function,
			TermTypeFormula termTypeFormula0) {
		if (termTypeFormula0 == null)
			throw new IllegalArgumentException();
		Constant2Impl constant = newConstant(function);
		constant.termType0 = null;
		constant.termTypeFormula0 = termTypeFormula0;
		constant.termType1 = null;
		return constant;
	}

	private TermImpl arg1 = null;
	
	private TermType termType0;
	private TermTypeFormula termTypeFormula0;
	private TermType termType1;

	private Constant2Impl() {
	}

	// ////////////////////////
	// Constant implementation
	// ////////////////////////

	public final Term apply(Term arg, TermRecycler recycler) {
		if (arg == null)
			throw new IllegalArgumentException();
		TermImpl lastArg = (TermImpl) arg;
		lastArg.referenceCount++;
		if (this.argumentCount == 0) {
			this.arg1 = lastArg;
			this.argumentCount = 1;
			return this;
		}
		TermImpl result = (TermImpl) this.function.function(this.arg1, lastArg);
		result.referenceCount++; // Protection against recycler
		lastArg.referenceCount--;
		if (lastArg.referenceCount == 0)
			recycler.breakTermDeeply(lastArg);
		this.arg1.referenceCount--;
		if (this.arg1.referenceCount == 0)
			recycler.breakTermDeeply(this.arg1);
		this.arg1 = null;
		result.referenceCount--;
		this.argumentCount = 0;
		this.breakIt();
		return result;
	}

	public final Term removeLastArg() {
		if (this.argumentCount == 0)
			return null;
		TermImpl arg = this.arg1;
		this.arg1 = null;
		if (arg != null)
			arg.referenceCount--;
		this.argumentCount = 0;
		return arg;
	}

	public final Term getArgument(int index) {
		if (index == 0 && this.argumentCount == 1)
			return this.arg1;
		throw new IllegalArgumentException();
	}

	public Constant cloneWithoutArgument() {
		return newConstant(this.function);
	}

	// ////////////////////
	// Term implementation
	// ////////////////////

	TermType assignType(List<Variable> boundVariableStack,
			List<TermType> boundVariableTypeStack,
			Map<Term, TermType> fromSharedTermToType) {
		if (this.argumentCount == 0) {
			if (this.termType0 != null)
				return this.termType0;
			if (this.termTypeFormula0 != null) {
				TermType variableTermType = TermTypeFactory
						.newVariableTermType();
				return this.termTypeFormula0.termTypeFormula(variableTermType);
			}
			throw new IllegalStateException();
		}
		if (this.argumentCount == 1) {
			if (this.termType1 != null)
				return this.termType1;
			TermType argTermType = this.arg1.assignType(boundVariableStack,
					boundVariableTypeStack, fromSharedTermToType);
			return TermTypeFactory.newAbstractionTermType(argTermType, 1);
		}
		throw new IllegalStateException();
	}

	// //////////////////////
	// ConstantImpl override
	// //////////////////////

	final void replaceArgument(int argumentIndex, TermImpl newValue) {
		if (argumentIndex == 0) {
			if (this.argumentCount < 1)
				throw new IllegalArgumentException();
			if (this.arg1 != null)
				this.arg1.referenceCount--;
			this.arg1 = newValue;
			if (newValue != null)
				newValue.referenceCount++;
			return;
		}
		throw new IllegalArgumentException();
	}

	// //////////////////
	// TermImpl override
	// //////////////////

	void breakIt() {
		if (TermImpl.ASSERTION && this.referenceCount != 0)
			throw new StillReferencedTermException(this);
		if (this.function == null)
			throw new BrokenTermException(this);
		this.function = null;
		if (this.arg1 != null) {
			this.arg1.referenceCount--;
			this.arg1 = null;
		}
		this.termType0 = null;
		this.termTypeFormula0 = null;
		this.termType1 = null;
		this.argumentCount = 0;
		reusableConstantStack.push(this);
		breakCount++;
	}
}
