/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl.coding;

import org.xmloperator.lambda3.impl.TermFactory;
import org.xmloperator.lambda3.model.Abstraction;
import org.xmloperator.lambda3.model.Application;
import org.xmloperator.lambda3.model.NumberCoding;
import org.xmloperator.lambda3.model.Term;
import org.xmloperator.lambda3.model.Variable;

/**
 * Implementation of NumberCoding as binary numerals.
 * 
 * <pre>
 * n = 2 * q + r
 * Term($n) = [isZeroOrOne:boolean, Term($q), r:boolean(0:F,1:K]
 * FunctionName ::= "isZero" | "succ" | "pred" | "add" | "mult" | "fromChurch" | "toChurch"
 * </pre>
 */
public final class BinaryNumberCoding implements NumberCoding {
	public static final NumberCoding NUMBER_CODING = new BinaryNumberCoding();

	/**
	 * Returns the value of the given binary numeral.
	 * 
	 * @param numeral
	 *            a binary numeral.
	 * @return the value of the given binary numeral or -1 if not a binary
	 *         numeral.
	 */
	static final int getValue(Abstraction numeral) {
		Term body = numeral.getBody();
		if (!body.isApplication())
			return -1;
		Application appliR = (Application) body;
		Term r = appliR.getArgument();
		if (!appliR.getFunction().isApplication())
			return -1;
		Application appliQ = (Application) appliR.getFunction();
		Term q = appliQ.getArgument();
		if (!appliQ.getFunction().isApplication())
			return -1;
		Application appliZ = (Application) appliQ.getFunction();
		Term z = appliZ.getArgument();
		if (appliZ.getFunction() != numeral.getBoundVariable())
			return -1;
		if (!z.isAbstraction() || !r.isAbstraction())
			return -1;
		Abstraction absZ = (Abstraction) z;
		Abstraction absR = (Abstraction) r;
		if (absZ.isK()) {
			if (absR.isK())
				return 1;
			if (absR.isF())
				return 0;
			return -1;
		}
		if (!q.isAbstraction())
			return 0;
		int value = getValue((Abstraction) q);
		if (value < 0)
			return -1;
		value += value;
		if (absR.isK())
			return value + 1;
		if (absR.isF())
			return value;
		return -1;
	}

	static final Variable newVariable() {
		return TermFactory.newVariable();
	}

	static final Variable newVariable(char nameChar) {
		return TermFactory.newVariable(nameChar);
	}

	/**
	 * Returns a new instance of $n with n < 2
	 * 
	 * @param rest
	 *            K for $1 or F for $0.
	 * @return a new instance of $n with n < 2
	 */
	static final Abstraction newSmallNumber(Term rest) {
		return Combinators.newTriplet(Combinators.newK(), Combinators.newI(),
				rest);
	}

	/**
	 * Returns a new instance of $n with n >= 2
	 * 
	 * @param quotient
	 *            $(n/2)
	 * @param rest
	 *            K for n%2==1 or F for n%2==0.
	 * @return a new instance of $n with n >= 2
	 */
	static final Abstraction newLargeNumber(Term quotient, Term rest) {
		return Combinators.newTriplet(Combinators.newF(), quotient, rest);
	}

	/**
	 * Returns a new instance of Term($n)
	 * 
	 * @param n
	 *            a positive or null number
	 * @return a new instance of Term($n)
	 */
	public static final Abstraction newNumber(int n) {
		if (n < 0)
			throw new IllegalArgumentException();
		return (n < 2 ? newSmallNumber(Combinators.newBoolean(n != 0))
				: newLargeNumber(newNumber(n / 2), Combinators
						.newBoolean(n % 2 != 0))).setCombinatorName("$" + n);
	}

	static final Abstraction newSelZ() {
		return Combinators.newOneOfN(2, 0);
	}

	static final Abstraction newSelQ() {
		return Combinators.newOneOfN(2, 1);
	}

	static final Abstraction newSelR() {
		return Combinators.newOneOfN(2, 2);
	}

	/**
	 * Returns a new instance of Term($isZero)
	 * 
	 * @return a new instance of Term($isZero)
	 */
	public static final Abstraction newIsZero() {
		Variable n = newVariable();
		return n.applyTo(newSelR()).applyTo(Combinators.newF()).applyTo(
				n.applyTo(newSelZ())).abstractOn(n)
				.setCombinatorName("$isZero");
	}

	/**
	 * Returns a new instance of Term($succ)
	 * 
	 * <pre>
	 * [z, $q, F] -> [z, $q, K]			(2q -> 2q+1)
	 * [K, I, K] -> [F, [K, I, K], F]	(1 -> 2)
	 * [F, $q, K] -> [F, $(q+1), F]]	(2q+1 -> 2q+2)
	 * 
	 * n=[z,q,r]
	 * YG, G=\tn.(n+1)[t]
	 * </pre>
	 * 
	 * @return a new instance of Term($succ)
	 */
	public static final Term newSucc() {
		Variable t = newVariable();
		Variable n = newVariable();
		Term nZ = n.applyTo(newSelZ());
		Term nQ = n.applyTo(newSelQ());
		Term nR = n.applyTo(newSelR());
		Term caseRF = Combinators.newTriplet(nZ, nQ, Combinators.newK());
		Term caseRK = nZ.applyTo(newNumber(2)).applyTo(
				newLargeNumber(t.applyTo(nQ), Combinators.newF()));
		Abstraction g = nR.applyTo(caseRK).applyTo(caseRF).abstractOn(n)
				.abstractOn(t).setCombinatorName("$succ");
		return Combinators.newY().applyTo(g);
	}

	/**
	 * Returns a new instance of Term($pred)
	 * 
	 * <pre>
	 * [z, $q, K] -> [z, $q, F]			(2q+1 -> 2q)
	 * [K, I, F] -> [K, I, F]			(0 -> 0)
	 * [F, $1, F] -> [K, I, K]			(2 -> 1)
	 * [F, $q, F] -> [F, $(q-1), K]]	(2q -> 2(q-1)+1)
	 * 
	 * n=[z,q,r]
	 * YG, G=\tn.(n-1)[t]
	 * </pre>
	 * 
	 * @return a new instance of Term($pred)
	 */
	public static final Term newPred() {
		Abstraction K = Combinators.newK();
		Abstraction F = Combinators.newF();
		Abstraction Z = newSelZ();
		Abstraction Q = newSelQ();
		Abstraction R = newSelR();
		Variable t = newVariable();
		Variable n = newVariable();
		Term nZ = n.applyTo(Z);
		Term nQ = n.applyTo(Q);
		Term nR = n.applyTo(R);
		Term caseRK = Combinators.newTriplet(nZ, nQ, F);
		Term caseRF = nZ.applyTo(n).applyTo(
				nQ.applyTo(Z).applyTo(newNumber(1)).applyTo(
						newLargeNumber(t.applyTo(nQ), K)));
		Abstraction g = nR.applyTo(caseRK).applyTo(caseRF).abstractOn(n)
				.abstractOn(t).setCombinatorName("$pred");
		return Combinators.newY().applyTo(g);
	}

	/**
	 * Returns a new instance of Term($add)
	 * 
	 * <pre>
	 * [K, I, F] + $n -> $n									(0+n -> n)
	 * [K, I, K] + $n -> $succ($n)							(1+n -> n+1)
	 * [F, $q1, F] + [F, $q2, F] -> [F, $q1+$q2, F]			(2q1+2q2 -> 2(q1+q2))
	 * [F, $q1, F] + [F, $q2, K] -> [F, $q1+$q2, K]			(2q1+2q2+1 -> 2(q1+q2)+1)
	 * [F, $q1, K] + [F, $q2, K] -> [F, $succ($q1+$q2), F]	(2q1+1+2q2+1 -> 2(q1+q2+1))
	 * 
	 * $n1 = [z1,$q1,r1], $n2 = [z2,$q2,r2]
	 * YG, G=\tn1n2.(n1+n2)[t]
	 * </pre>
	 * 
	 * @return a new instance of Term($add)
	 */
	public static final Term newAdd() {
		Abstraction K = Combinators.newK();
		Abstraction F = Combinators.newF();
		Abstraction Z = newSelZ();
		Abstraction Q = newSelQ();
		Abstraction R = newSelR();
		Variable t = newVariable();
		Variable n1 = newVariable();
		Variable n2 = newVariable();
		Term n1Z = n1.applyTo(Z);
		Term n2Z = n2.applyTo(Z);
		Term n1Q = n1.applyTo(Q);
		Term n2Q = n2.applyTo(Q);
		Term n1R = n1.applyTo(R);
		Term n2R = n2.applyTo(R);
		Term succ = newSucc();
		Term caseN1Z = n1R.applyTo(succ.applyTo(n2)).applyTo(n2);
		Term caseN2Z = n2R.applyTo(succ.applyTo(n1)).applyTo(n1);
		Term tQ1Q2 = t.applyTo(n1Q).applyTo(n2Q);
		Term caseRFF = newLargeNumber(tQ1Q2, F);
		Term caseRFK = newLargeNumber(tQ1Q2, K);
		Term caseRKF = caseRFK;
		Term caseRKK = newLargeNumber(succ.applyTo(tQ1Q2), F);
		Abstraction g = n1Z.applyTo(caseN1Z).applyTo(
				n2Z.applyTo(caseN2Z)
						.applyTo(
								n1R.applyTo(
										n2R.applyTo(caseRKK).applyTo(caseRFK))
										.applyTo(
												n2R.applyTo(caseRKF).applyTo(
														caseRFF)))).abstractOn(
				n2).abstractOn(n1).abstractOn(t).setCombinatorName("$add");
		return Combinators.newY().applyTo(g);
	}

	/**
	 * Returns a new instance of Term($mult)
	 * 
	 * <pre>
	 * [K, I, F] * $n -> $0						(0*n -> 0)
	 * [K, I, K] * $n -> $n						(1*n -> n)
	 * [F, $q, F] * $n -> [F, $q*$n, F]			(2q*n -> 2(q*n))
	 * [F, $q, K] * $n -> [F, $q*$n, F] + $n	((2q+1)*n -> 2(q*n)+n)
	 * 
	 * $n1 = [z1,$q1,r1]
	 * YG, G=\tn1n2.(n1*n2)[t]
	 * </pre>
	 * 
	 * @return a new instance of Term($mult)
	 */
	public static final Term newMult() {
		Abstraction F = Combinators.newF();
		Variable t = newVariable(); // ($n1,$n2) -> $(n1*n2)
		Variable n1 = newVariable();
		Variable n2 = newVariable();
		Term z1 = n1.applyTo(newSelZ());
		Term q1 = n1.applyTo(newSelQ());
		Term r1 = n1.applyTo(newSelR());
		Term tq1n2 = t.applyTo(q1).applyTo(n2);
		Term caseF = newLargeNumber(tq1n2, F);
		Term caseK = newAdd().applyTo(caseF).applyTo(n2);
		Abstraction g = z1.applyTo(r1.applyTo(n2).applyTo(newSmallNumber(F)))
				.applyTo(r1.applyTo(caseK).applyTo(caseF)).abstractOn(n2)
				.abstractOn(n1).abstractOn(t).setCombinatorName("$mult");
		return Combinators.newY().applyTo(g);
	}

	/**
	 * Returns a new instance of Church to Binary numeral transformer.
	 * 
	 * <pre>
	 * #n -> t(#n)
	 * #0 -> [K, I, F]
	 * #1 -> [K, I, K]
	 * [#q,r] -> [F, t(#q), r]
	 * </pre>
	 * 
	 * @return a new instance of Church to Binary numeral transformer.
	 */
	public static final Term newFromChurch() {
		Variable t = newVariable(); // #n -> t(#n)
		Variable p = newVariable(); // [#q,r]
		Term pK = p.applyTo(Combinators.newK());
		Term pF = p.applyTo(Combinators.newF());
		// p -> t(#n)
		Abstraction f = ChurchNumberCoding.newIsZero().applyTo(pK).applyTo(
				newSmallNumber(pF)).applyTo(newLargeNumber(t.applyTo(pK), pF))
				.abstractOn(p);
		Variable n = newVariable(); // #n
		Abstraction g = f.applyTo(ChurchNumberCoding.newDivByTwo().applyTo(n))
				.abstractOn(n).abstractOn(t).setCombinatorName("$fromChurch");
		return Combinators.newY().applyTo(g);
	}

	/**
	 * Returns a new instance of Binary to Church numeral transformer.
	 * 
	 * <pre>
	 * $n -> #n
	 * $0 -> #0
	 * $1 -> #1
	 * [F, $q, r] -> r ? #2q+1 : #2q
	 * </pre>
	 * 
	 * @return a new instance of Binary to Church numeral transformer.
	 */
	public static final Term newToChurch() {
		Variable t = newVariable(); // $n -> t($n)
		Variable n = newVariable(); // $n
		Term nZ = n.applyTo(newSelZ());
		Term nQ = n.applyTo(newSelQ());
		Term nR = n.applyTo(newSelR());
		Abstraction C0 = ChurchNumberCoding.newZero();
		Abstraction C1 = ChurchNumberCoding.newNumber(1);
		Term caseZK = nR.applyTo(C1).applyTo(C0);
		Abstraction Csucc = ChurchNumberCoding.newSucc();
		Abstraction Cadd = ChurchNumberCoding.newAdd();
		Term tq = t.applyTo(nQ);
		Term tq2 = Cadd.applyTo(tq).applyTo(tq);
		Term caseZF = nR.applyTo(Csucc.applyTo(tq2)).applyTo(tq2);
		Abstraction g = nZ.applyTo(caseZK).applyTo(caseZF).abstractOn(n)
				.abstractOn(t).setCombinatorName("$toChurch");
		return Combinators.newY().applyTo(g);
	}

	/**
	 * Returns a new instance of the Syracuse conjecture operator.
	 * 
	 * <pre>
	 * [K,I,F] -> [K,I,F]				(0 -> 0)
	 * [K,I,K] -> [K,I,K]				(1 -> 1)
	 * [F,p,F] -> p						(2p -> p)
	 * [F,p,K] -> [F,p,K] + $succ(p)	(2p+1 -> 3p+2)
	 * </pre>
	 * 
	 * @return a new instance of the Syracuse conjecture operator.
	 */
	public static final Abstraction newSyracuse() {
		Variable n = newVariable();
		Term nZ = n.applyTo(newSelZ());
		Term nQ = n.applyTo(newSelQ());
		Term nR = n.applyTo(newSelR());
		return nZ.applyTo(n).applyTo(
				nR.applyTo(newAdd().applyTo(n).applyTo(newSucc().applyTo(nQ)))
						.applyTo(nQ)).abstractOn(n).setCombinatorName(
				"$syracuse");
	}

	/**
	 * Returns a new instance of the operator that gives the step count of the
	 * Syracuse sequence starting at a given number.
	 * 
	 * <pre>
	 * stepCount := 0
	 * while Z(n) != 0
	 * 		stepCount += 1
	 * 		n := syracuse(n)
	 * end while
	 * return stepCount
	 * </pre>
	 * 
	 * @return a new instance of the operator that gives the step count of the
	 *         Syracuse sequence starting at a given number.
	 */
	public static final Term newSyracuseStepCount() {
		Variable t = newVariable();
		Variable n = newVariable();
		Abstraction g = n.applyTo(newSelZ()).applyTo(
				ChurchNumberCoding.newZero()).applyTo(
				ChurchNumberCoding.newSucc().applyTo(
						t.applyTo(newSyracuse().applyTo(n)))).abstractOn(n)
				.abstractOn(t).setCombinatorName("$syracuseStepCount");
		return Combinators.newY().applyTo(g);
	}

	// ////////////////////////////
	// NumberCoding implementation
	// ////////////////////////////

	public char getIntroductionChar() {
		return NumberCoding.BINARY;
	}

	public Term codeNumber(boolean isSigned, int number) {
		if (isSigned)
			throw new IllegalArgumentException();
		return newNumber(number);
	}

	public Term codeFunction(String functionName) {
		if (functionName.equals(ISZERO))
			return newIsZero();
		if (functionName.equals(SUCC))
			return newSucc();
		if (functionName.equals(PRED))
			return newPred();
		if (functionName.equals(ADD))
			return newAdd();
		if (functionName.equals(MULT))
			return newMult();
		if (functionName.equals(FROM_CHURCH))
			return newFromChurch();
		if (functionName.equals(TO_CHURCH))
			return newToChurch();
		if (functionName.equals("syracuse"))
			return newSyracuse();
		if (functionName.equals("syracuseStepCount"))
			return newSyracuseStepCount();
		return null;
	}
}
