/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.model;

import org.xmloperator.lambda3.exception.term.NotARedexException;
import org.xmloperator.lambda3.exception.term.StillReferencedTermException;

/**
 * Tools for Term reducing.
 */
public interface TermReducer extends TermRecycler {

	/**
	 * Returns the beta or delta reduction of a redex. The redex must be not
	 * referenced.
	 * <p>
	 * The reduction is a beta reduction if the function is an Abstraction. The
	 * reduction is a delta reduction if the function is a Constant.
	 * </p>
	 * 
	 * @param redex
	 *            a redex to reduce.
	 * @return the beta or delta reduction of the redex.
	 * @throws NotARedexException
	 *             if the gien Application is not a redex.
	 * @throws StillReferencedTermException
	 *             if the redex is referenced.
	 */
	public Term betaDeltaReduce(Application redex);

	/**
	 * Reduces the given redex within the given root Term. The redex must be
	 * referenced once or more, exclusively within the root Term. The redex
	 * cannot be the root Term.
	 * 
	 * @param rootTerm
	 *            the root Term.
	 * @param redex
	 *            a redex within the root Term.
	 * @throws NotARedexException
	 *             if the given Application is not a redex.
	 * @throws IllegalArgumentException
	 *             if the redex is referenced outside this Term or is this Term.
	 */
	public void betaDeltaReduce(Term rootTerm, Application redex);

	/**
	 * Returns the leftermost-outermost redex of the given root Term or null if
	 * this Term doesn't contain any redex.
	 * 
	 * @param Term
	 *            a root Term.
	 * @param isLazy
	 *            if <tt>true</tt> then doesn't go within abstractions.
	 * @return the leftermost-outermost redex of the given root Term or null if
	 *         the root Term doesn't contain any redex.
	 */
	public Application leftermostOutermostRedex(Term rootTerm,
			final boolean isLazy);

	/**
	 * Returns the normal form of a given root Term, if possible.
	 * <p>
	 * The root Term must be closed and not referenced.
	 * </p>
	 * 
	 * @param rootTerm
	 *            a root Term.
	 * @param isLazy
	 *            <tt>true</tt> for a lazy reduction, <tt>false</tt> for a
	 *            strong one.
	 * @param informantReduction
	 *            an optional InformantReduction
	 * @return the normal form of the root Term, if possible.
	 * @throws StillReferencedTermException
	 *             if the root Term is referenced.
	 */
	public Term normalForm(Term rootTerm, boolean isLazy,
			InformantReduction informantReduction);

	/**
	 * Returns the normal form of a given root Term.
	 * <p>
	 * The root Term must have a normal form, be closed and not referenced.
	 * </p>
	 * 
	 * @param rootTerm
	 *            a root Term.
	 * @return the normal form of the root Term.
	 * @throws StillReferencedTermException
	 *             if the root Term is referenced.
	 */
	public Term normalForm(Term rootTerm);
}
